<?php

/**
 * WC_Zeamster_Functions Main implementation of GW's card processing
 *
 * @category Class
 * @package  Woocommerce
 * @author   None <none@fake.com>
 * @license  http://www.gnu.org/copyleft/gpl.html GNU General Public License
 * @link     http://www.none.com/
 */
class WC_Zeamster_Functions
{
	protected $options;
	public $_logger;

	public function __construct() {
		$this->options = get_option("woocommerce_". GW_ZP_ID ."_settings");

		// Endpoints to be used in the Plugin
		$this->endpoints = array (
			"process_transaction" => "/v2/transactions", // POST
			"refund_transaction"  => "/v2/transactions", // POST
			"void_transaction" 	  => "/v2/transactions/{id}", // PUT
			"partial_reversal" 	  => "/v2/transactions/{id}", // PUT
			"auth_complete"	   	  => "/v2/transactions/{id}", // PUT
			"get_transaction" 	  => "/v2/transactions/{id}?expand=is_voidable,is_refundable,transaction_level3", // GET
			"get_account_vault"   => "/v2/accountvaults/{id}", // GET
			"update_level_3"	  => "/v2/transactionlevel3s"  //POST
		);

		$this->_logger = new ZP_Logger();
		$this->_logger->setFebug($this->options["debug"] === "yes" ? true : false);
	}

	/** 
	 * Get Transaction from the API. 
	 * 
	 * This will be performed prior to refunds/voids/partial reversals 
	 * as WooCommerce does not know the settled status of a transaction.
	 * 
	 * @param String $transaction_id | Gateway Transaction ID of Original Transaction
	 */
	public function getTransaction ( $transaction_id ) {
		$endpoint = $this->endpoints["get_transaction"];
		$endpoint = str_replace('{id}', $transaction_id, $endpoint);
		$endpoint = $this->getApiUrl() . $endpoint;

		$result = $this->makeApiCall($endpoint, "GET");
		return $result;
	}

	/**
	 * Process sale using a stored account vault record
	 * 
	 * @param WC_Payment_Token $token | Instance of WC_Payment_Token which stored payment token 
	 * @param WC_Order $order | Instance of WC_Order which has the order information
	 * 
	 * @return Array $result | API Response from Zeamster
	 */
	public function processSaleFromAv( $token, $order ) { 
		$endpoint = $this->endpoints["process_transaction"];
		$endpoint = $this->getApiUrl() . $endpoint;
		$payload = $this->getAvTransactionData( $token, $order );
		$result = $this->makeApiCall($endpoint, "POST", $payload);

		return $result;
	}

	/** 
	 * Get Level 3 Data for applicable transactions and add them to the transaction
	 * 
	 * @param WC_Order $order | Instance of WC_Order which has the order information
	 * @param String $transaction_id | Transaction ID to query for Level 3
	 */
	public function processLevelThreeData( WC_Order $order, int $order_id, string $transaction_id ) : bool
	{ 
		$result = $this->getTransaction( $transaction_id );
		
		if ( $result [ "transaction" ][ "transaction_level3" ] != null ) { 
			$lineItems = [];

			foreach ( $order->get_items() as $item_id => $item ) {
				$itemArray = [
					"description" => substr($item->get_name(), 0 , 25),
					"product_code" => $item->get_product_id(),
					"quantity" => $item->get_quantity(),
					"tax_amount" => number_format($item->get_subtotal_tax(), 2, '.', ','),
					"tax_rate" => $this->getItemTaxRate($item),
					"unit_code" => "EA",
					"unit_cost" => number_format($item->get_subtotal(), 2, '.', ','),
					"commodity_code" => $result [ "transaction" ][ "transaction_level3" ][ "level3_data" ][ "summary_commodity_code" ],
					"tax_type_applied" => "00",
					"tax_type_id" => "00"
				];

				array_push($lineItems, $itemArray);				
			}

			$orderDate = $order->get_date_created()->date;
			$date = new DateTime($orderDate);
			
			$payload = [
				"transactionlevel3" =>  [
					"transaction_id" => $transaction_id,
					"level3_data" => [
						"tax_exempt" => "0",
						"sales_tax" => $order->get_total_tax('edit'),
						"national_tax" => $result [ "transaction" ][ "transaction_level3" ][ "level3_data" ][ "national_tax" ],
						"merchant_vat_registration" => $result [ "transaction" ][ "transaction_level3" ][ "level3_data" ][ "merchant_vat_registration" ],
						"summary_commodity_code" => $result [ "transaction" ][ "transaction_level3" ][ "level3_data" ][ "summary_commodity_code" ],
						"freight_amount" => $order->get_shipping_total(),
						"duty_amount" => "0",
						"shipto_zip_code" => $order->get_shipping_postcode(),
						"shipfrom_zip_code" => $result [ "transaction" ][ "transaction_level3" ][ "level3_data" ][ "shipfrom_zip_code" ],
						"destination_country_code" => "840",
						"unique_vat_ref_number" => str_pad($order->get_id(), 17, 0, STR_PAD_LEFT),
						"order_date" => $date->format("ymd"),
						"tax_amount" => $order->get_total_tax('edit'),
						"tax_rate" => "0",
						"line_items" => $lineItems 
					]
				]
			];

			$this->_logger->info(json_encode("Line: ". __LINE__ .": Level 3 Payload ". $payload ));

			$endpoint = $this->endpoints["update_level_3"];
			$endpoint = $this->getApiUrl() . $endpoint;
			$result = $this->makeApiCall($endpoint, "POST", $payload);

			$this->_logger->info("Line: ". __LINE__ .": Level 3 API Response ". json_encode($result));

			if ($result[ "transactionlevel3" ]) { 
				return true;						
			} else { 
				return false;
			}
		}	

		return false;
	}

	/**
	 * Refund a transaction from within WooCommerce
	 * 
	 * @param String $amount  | Amount of refund - Partial refunds allowed
	 * @param String $transaction_id | Gateway Transaction ID
	 * @return Array $result | API response in Array Format
	 */
	public function refundTransaction( $amount, $transaction_id ) { 
		$endpoint = $this->endpoints["refund_transaction"];
		$endpoint = $this->getApiUrl() . $endpoint;

		$payload = [
			"transaction" => [
				"action" => "refund",
				"payment_method" => "cc",
				"previous_transaction_id" => $transaction_id,
				"transaction_amount" => $amount
			]
		];

		$this->_logger->info(json_encode("Line: ". __LINE__ .": Refund Payload ". $payload ));

		$result = $this->makeApiCall($endpoint, "POST", $payload);

		$this->_logger->info("Line: ". __LINE__ .": Refund Response ". json_encode($result));

		return $result;
	}

	/**
	 * Void a transaction from within WooCommerce
	 * 
	 * @param String $transaction_id | Gateway Transaction ID
	 * @return Array $result | Array of Void Response
	 */
	public function voidTransaction( $transaction_id ) { 
		$endpoint = $this->endpoints["void_transaction"];
		$endpoint = str_replace('{id}', $transaction_id, $endpoint);
		$endpoint = $this->getApiUrl() . $endpoint;

		$payload = [
			"transaction" => [
				"action" => "void"
			]
		];

		$this->_logger->info(json_encode("Line: ". __LINE__ .": Void Endpoint ". $endpoint ));

		$result = $this->makeApiCall($endpoint, "PUT", $payload);

		$this->_logger->info(json_encode("Line: ". __LINE__ .": Void Response ". $result ));

		return $result;
	}

	/**
	 * Partially Reverse a transaction from within WooCommerce
	 * which is still yet to be settled in the API 
	 * 
	 * @param String $transaction_id  | Gateway Transaction ID
	 * @param String #amountRemaining | Amount remaining after refund. Passed as "transaction_amount" in request
	 * @return Array $result | Array of Partial Reversal Response
	 */
	public function partialReversalTransaction( $transaction_id, $amountRemaining ) {
		$endpoint = $this->endpoints["partial_reversal"];
		$endpoint = str_replace( '{id}', $transaction_id, $endpoint) ;
		$endpoint = $this->getApiUrl() . $endpoint;

		$payload = [
			"transaction" => [
				"action" => "partialreversal",
				"transaction_amount" => $amountRemaining // will change the amount that needs to be authed.
			]
		];

		$this->_logger->info(json_encode("Line: ". __LINE__ .": Partial Reversal Payload ". $payload ));

		$result = $this->makeApiCall($endpoint, "PUT", $payload);

		$this->_logger->info(json_encode("Line: ". __LINE__ .": Partial Reversal Response ". $result ));

		return $result;
	}	

	/**
	 * Capture a transaction when processed with Authonly
	 * 
	 * @param WC_Order $order | Instance of WC_Order - Order Object
	 * @param Int $order_id | Order ID
	 * @param String $transaction_id | Transaction ID
	 * @param Float $amount | Transaction Amount
	 * 
	 * @return Array $response | Transaction API Response from the 
	 */
	public function authCompleteTransaction( $order, $order_id, $transaction_id, $amount ) {
		$endpoint = $this->endpoints["auth_complete"];
		$endpoint = str_replace('{id}', $transaction_id, $endpoint);
		$endpoint = $this->getApiUrl() . $endpoint;

		$payload = [
			"transaction" => [
				"action" => "authcomplete", 
				"transaction_amount" => $amount
			]
		];

		$this->_logger->info(json_encode("Line: ". __LINE__ .": Auth Complete Payload ". $payload ));
				
		$response = $this->makeApiCall($endpoint, "PUT", $payload);

		$this->_logger->info(json_encode("Line: ". __LINE__ .": Auth Complete Response ". $response ));

		return $response;

	}

	/**
	 * Get account vault record from Zeamster API 
	 * 
	 * @param String $account_vault_id | The ID of the account vault to be returned
	 * 
	 * @return Array $response | Account Vault Response from the Zeamster API.
	 */
	public function getAccountVaultRecord( $account_vault_id ) {
		$endpoint = $this->endpoints["get_account_vault"];
		$endpoint = str_replace("{id}", $account_vault_id, $endpoint);
		$endpoint = $this->getApiUrl() . $endpoint;
		$response = $this->makeApiCall($endpoint, "GET");

		return $response;
	}

	/**
	 * @Todo - Add in Regex Validation and other restraints
	 * ---
	 * Check whether Environment is set to Sandbox or Production and then
	 * check the settings associated with that environment. If all the settings
	 * are input correctly, return true otherwise false
	 * 
	 * @return Boolean true|false
	 */
	public function checkZpConfig() {
		$environment = $this->getEnvironment();
		if ( !$environment ) return false;

		$host = $this->getHost();
		if ( !$host ) return false;

		$hppId = $this->getHppId();
		if ( !$hppId ) return false;

		$encryptionKey = $this->getEncryptionKey();
		if ( !$encryptionKey ) return false;

		$userId = $this->getUserId();
		if ( !$userId ) return false;

		$userApiKey = $this->getUserApiKey();
		if ( !$userApiKey ) return false;

		return true;
	}

	/**
	 * Helper method to get the currently configured Environment
	 * 
	 * @return String|Bool $environment|false
	 */
	public function getEnvironment() { 
		$environment = $this->options["environment"];

		switch (trim($environment)) { 
			case "": return false;
			case "sandbox": return "sandbox";
			case "production": return "production";
			default: return false;
		}

		return false;
	}

	/**
	 * Helper method to get the currently configured API Host
	 * 
	 * @return String|Bool $host|false
	 */
	public function getHost() { 
		if ( trim($this->options['environment']) === "production" ) { 
			$host = $this->options["production_host_domain"];
		}

		if ( trim($this->options['environment']) === "sandbox" ) { 
			$host = $this->options["sandbox_host_domain"];
		}

		if ( !isset($host) || trim($host) === "" ) { 
			return false;
		}

		return $host;
	}

	/**
	 * Helper method to get the currently configured 
	 * HPP Encryption Key
	 * 
	 * @return String|Bool $key|false
	 */
	public function getEncryptionKey() { 
		if ( trim($this->options['environment']) === "production" ) { 
			$key = $this->options["production_hpp_encryption_key"];
		}

		if ( trim($this->options['environment']) === "sandbox" ) { 
			$key = $this->options["sandbox_hpp_encryption_key"];
		}

		if ( !isset($key) || trim($key) === "" ) { 
			return false;
		}

		if ( strlen( $key ) !== 32 )  { 
			return false;
		}

		return $key;
	}

	/**
	 * Helper method to get the FQDN from the current Host Settings
	 * 
	 * @return String|Bool $url|false
	 */
	public function getApiUrl() { 
		if ( $this->options['environment'] == "sandbox" ) { 
			return "https://api.sandbox.". $this->options["sandbox_host_domain"];
		} 

		if ( $this->options['environment'] == "production" ) {
			return "https://api.". $this->options["production_host_domain"];
		}

		return false;
	}

	/**
	 * Helper method to get the full HPP url based on Environment, Host and HPP ID
	 * 
	 * @param  String $data
	 * @return String $url
	 */
	public function getHppUrl ( string $data ) { 
		if ( trim($this->options['environment']) === "production" ) { 
			$url = 'https://api.'. $this->options['production_host_domain'] . '/hostedpaymentpage?id=' . $this->options['production_hpp_id'] . '&expand=saved_account_vault_id&data=' . $data;
		}

		if ( trim($this->options['environment']) === "sandbox" ) { 
			$url = 'https://api.sandbox.'. $this->options['sandbox_host_domain'] . '/hostedpaymentpage?id=' . $this->options['sandbox_hpp_id'] . '&expand=saved_account_vault_id&data=' . $data;
		}

		$this->_logger->info("HPP URL being generated is: " . $url);

		return $url;
	}

	/**
	 * @param WC_Payment_Token $token | Instance of WC_Payment_Token which stores the Account Vault ID
	 * @param WC_Order $order | Instance of WC_Order for processing
	 * 
	 * @return Array $payload | Payload to be sent to Zeamster
 	 */
	private function getAvTransactionData( $token, $order ) { 
		$zpfn = new WC_Zeamster_Functions;
		// Get Transaction Overrides
		$overrides = $zpfn->getOverrideOptions();

		$payload = [
			"transaction" => [
				"action" => $this->options[ 'transactiontype' ],
				"payment_method" => "cc",
				"account_vault_id" => $token->get_token(),
				"transaction_amount" => $order->get_total(),
				"order_num" => str_pad( $order->get_order_number(), 17, "0", STR_PAD_LEFT ),
				"po_number" => str_pad( $order->get_order_number(), 17, "0", STR_PAD_LEFT ),
				"billing_street" => $order->get_billing_address_1(),
				"billing_city" => $order->get_billing_city(),
				"billing_state" => $order->get_billing_state(),
				"billing_zip" => $order->get_billing_postcode(),
// 				"billing_phone" => $order->get_billing_phone(),
				"allow_partial_authorization_override" => false,
				"auto_decline_street_override" => $overrides[ 'street' ],
				"auto_decline_zip_override" => $overrides[ 'zip' ],	
			],
		];

		$this->_logger->info("Line: ". __LINE__ ."The payload for the account vault transaction is: " . json_encode($payload));

		return $payload;
	}

	/**
     * Build the payment request
     *
     * @param String   $orderID              Order ID
     * @param String   $creditCardNumber     Credit card number
     * @param String   $creditCardExpiration Credit Card Expiration Date
     * @param WC_Order $order                The actual order
     *
     * @return Array
     */
    public function getHppRequestData( int $orderID, WC_Order $order, string $saveCard ) {
		$overrides = $this->getOverrideOptions();
		$args = array(
			"id" => $this->getHppId(),
			"min_payment_amount" => $order->get_total(),
			"max_payment_amount" => $order->get_total(),
			"field_configuration" => [
				"body" => [
					"fields" => [
						// Overrides
						[
							"id" => "allow_partial_authorization_override", 
							"value" => false, 
							"readonly" => true,
							"visible" => false,
						], 
						[
							"id" => "auto_decline_cvv_override", 
							"value" => $overrides[ 'cvv' ],
							"readonly" => true,
							"visible" => false,
						], 
						[
							"id" => "auto_decline_street_override", 
							"value" => $overrides[ 'street' ],
							"readonly" => true,
							"visible" => false,
						], 
						[
							"id" => "auto_decline_zip_override", 
							"value" => $overrides[ 'zip' ],
							"readonly" => true,
							"visible" => false,
						], 
						// Transaction Details
						[
							"id" => "action",
							"value" => $this->options[ 'transactiontype' ],
							"visible" => false,
						],
						[
							"id" => "transaction_amount", 
							"value" => $order->get_total(),
							"label" => "Amount",
							"readonly" => true,
							"visible" => true,
						], 
						[
							"id" => "order_num",
							"value" => str_pad( $orderID, 17, "0", STR_PAD_LEFT ),
							"label" => "Order Number",
							"readonly" => true,
							"visible" => true,
						],
						[
							"id" => "po_number",
							"value" => str_pad( $orderID, 17, "0", STR_PAD_LEFT ),
							"label" => "PO Number",
							"readonly" => true,
							"visible" => false,
						],
						// Customer Details						
						[
							"id" => "billing_street",
							"value" => $order->get_billing_address_1(),
							"label" => "Billing Street",
							"readonly" => true,
							"visible" => true,
						],
						[
							"id" => "billing_city",
							"value" => $order->get_billing_city(),
							"label" => "Billing City",
							"readonly" => true,
							"visible" => true,
						],
						[
							"id" => "billing_state",
							"value" => $order->get_billing_state(),
							"label" => "Billing State",
							"readonly" => true,
							"visible" => true,
						],
						[
							"id" => "billing_zip",
							"value" => str_replace('-', '', str_replace(' ', '', $order->get_billing_postcode())),
							"label" => "Billing Zip Code",
							"readonly" => true,
							"visible" => true,
						],
// 						[
// 							"id" => "billing_phone",
// 							"value" => $order->get_billing_phone(),
// 							"label" => "Billing Phone",
// 							"readonly" => true,
// 							"visible" => true,
// 						],
						[
							"id" => "save_account",
							"value" => $saveCard === "false" ? 0 : 1,
							"visible" => true
						]
					]
				]
			]
		);

		$this->_logger->info(json_encode($args));
        $this->_logger->info( 'gateway request [orderID:'. $orderID .']: ' . json_encode( $args ) );
		
        return $args;
	}

	/**
	 * Encode the data to append to the HPP URL
	 * 
	 * @param  Array  $args | Array of custom data
	 * @return String $encodedData | Encoded data string based on Encryption Key
	 */
	public function encodeRequestData( array $args ) { 
		$encryption_key = $this->getEncryptionKey();

		// Convert to JSON string
		$data = json_encode( $args );
		$salt = openssl_random_pseudo_bytes( 8 );
		$salted = '';
		$dx = '';
		
		while ( strlen( $salted ) < 48 ) {
			$dx = md5( $dx . $encryption_key . $salt, true );
			$salted .= $dx;
		}

		$key = substr( $salted, 0, 32 );
		$iv = substr( $salted, 32, 16 );

		// Encrypt the JSON object
		$encryptedString = openssl_encrypt( $data, 'aes-256-cbc', $key, true, $iv );

		return urlencode( base64_encode( "Salted__" . $salt . $encryptedString ) );
	}

	/** 
	 * Get Allowed postback URL based on Current Host and Environment
	 * 
	 * @return String|Bool $url|false
	 */
	public function getAllowedPostbackUrl() { 
		if (trim($this->options['environment']) === "sandbox") { 
			$allowedUrl = "https://api.sandbox." . $this->options['sandbox_host_domain'];
			$this->_logger->info("Line ". __LINE__ .": Allowed Postback URL is: " . $allowedUrl);
			return $allowedUrl;
		}
		
		if (trim($this->options['environment']) === "production") { 
			$allowedUrl = "https://api." . $this->options['production_host_domain'];
			$this->_logger->info("Line ". __LINE__ .": Allowed Postback URL is: " . $allowedUrl);
			return $allowedUrl;
		}

		return false;
	}

	/**
	 * Check response from API for Void, Refund, Partial Reversal
	 * 
	 * @param Array $response | API Response in Array Format
	 * @param WC_Order $order | Instance of WooCommerce Order
	 * @param Integer $order_id | WooCommerce Order ID
	 * @param String $amount | Amount being refunded
	 * @param Integer $expectedResponseCode
	 * 				  Void = 201 - Successful Void
	 * 				  PartialReversal = 101 - Successful Partial Reversal
	 *  		   	  Refund = 111 - Successful Refund
	 * 
	 * @return Bool true|false
	 */
	public function validateApiRefundResponse( $response, $order, $order_id, $amount, $expectedResponseCode ) { 
		// Check for error when processing refund
		if ( !in_array( $expectedResponseCode, [ 101, 111, 201 ], true ) ) { 
			$this->_logger->info( 'Error Occured when attempting to process refund: ' . json_encode( $response ) );
			wc_add_notice( __( 'Unable to confirm refund was successful', GW_ID ) );
			return false;
		}
	
		// Check for valid transaction
		if ( !is_wp_error( $response ) && $response[ 'transaction' ][ 'status_id' ] === $expectedResponseCode ) { 
			$this->addOrderRefundNote( $order, $order_id, $amount, $response );
			return true;
		} else {
			$this->_logger->info( "FAILURE: Unable to refund funds for Transaction: " . $response[ 'transaction' ][ 'id' ] );
			$this->orderRefundErrorNotice( $response );
			return false;
		}
	}

	/**
	 * @param Array | $response - Response from the API for a Capture Transaction
	 * @param WC_Order | $order - Instance of WC_Order 
	 * @param Int | $order_id - ID of the associated order being handled
	 * @param Float | $amount - Amount of the transaction being captured (In USD)
	 * @param Array | $expectedResponseCode - Array list of expected response codes
	 * 
	 * 101 = Approved CC Trx
	 * 111 = Approved CC Refund Trx
	 * 201 = Approved CC Void Trx
	 * 
	 * @return Boolean | [true,false] - Whether the transaction was successful or not
	 */
	public function validateApiCaptureResponse( $response, $order, $order_id, $amount, $expectedResponseCode ) { 
		if ( $response && isset( $response[ 'status' ] ) && in_array($response[ 'status' ], [400, 401, 403, 405, 409])) { 
			$args = [
				"status" => "error",
				"status_code" => $response[ 'status' ],
				"message" => $response[ 'message' ]
			];

			$this->_logger->info( 'Error details are: '. json_encode( $args ) );

			return $args;
		}

		// Check for Error in Trx
		if ( !in_array( $expectedResponseCode, [ 101, 111, 201 ], true ) ) { 
			// Ensure correct status_id
			if ( $response[ 'transaction' ][ 'type_id' ] === 20 ) {  // Type_ID of 20 = sale
				$this->_logger->info(json_encode($response));
				$this->_logger->info('Unable to complete transaction');
				return false;
			} else { 
				$this->_logger->info(json_encode($response));
				$this->_logger->info('Unable to complete transaction');
				return false;
			}

			return false;
		}

		// Check for correct Response Code + Non WP Error
		if ( !is_wp_error( $response ) && $response[ 'transaction' ][ 'status_id' ] === $expectedResponseCode ) { 
			// If sale return true
			if ( $response[ 'transaction' ][ 'type_id' ] === 20 ) { 
				return true;
			// If Void or Refund - Add Note to order, return true
			} else {
				$this->addOrderRefundNote( $order, $order_id, $amount, $response );
				return true;
			}
		} else {
			// Add Refund Error Note to Order + Return False 
			$this->_logger->info( "FAILURE: Unable to capture funds for Transaction: " . $response[ 'transaction' ][ 'id' ] );
			$this->orderRefundErrorNotice( $response );
			return false;
		}
	}

	/** 
	 * Add a refund message to an order after refunding all or partial funds
	 * 
	 * @param WC_Order $order | Instance of WooCommerce Order
	 * @param Integer $order_id
	 * @param String $amount
	 * @param Array $response
	 * 
	 * @return void
	 */
	private function addOrderRefundNote( $order, $order_id, $amount, $response ) { 
		// Add helpful notes
		$note = "Refund Processed:\n";
		$note.= "Transaction ID: ".sanitize_text_field( $response[ 'transaction' ][ 'id' ] )."\n";
		$note.= "Amount: ".sanitize_text_field($amount)."\n";
		$order->add_order_note( __( $note, 'zeamster' ) );

		$this->_logger->info( 'Gateway Response for Refund: '. json_encode( $response ) );

		// Add Status ID to DB
		update_post_meta(
			$order_id,
			'_payment_gateway_status',
			$response['transaction']['status_id']
		);

		$order->save();
	}

	private function orderRefundErrorNotice( $response ) { 
		global $woocommerce;
		$this->_logger->info( 
			'GW Response: '. json_encode($response) 
		);

		return false;
	}

	/**
	 * Get and compile array of overrides based on settings
	 * 
	 * @return Array $args | Array of Overrides
	 */
	private function getOverrideOptions() { 
		$args = array(
			"cvv" => $this->options["auto_decline_cvv"] == "yes" ? true : false,
			"street" => $this->options["auto_decline_street"] == "yes" ? true : false,
			"zip" => $this->options["auto_decline_zip"] == "yes" ? true : false,
		);

		return $args;
	}
 
	/** 
	 * Get HPP ID based on the currently configured environmet
	 * 
	 * @return String|Bool $id|false
	 */
	private function getHppId() { 
		if (trim($this->options['environment']) === "sandbox" && trim($this->options['sandbox_hpp_id']) !== "") { 
			$hppId = $this->options['sandbox_hpp_id'];
		}

		if (trim($this->options['environment']) === "production" && trim($this->options['production_hpp_id']) !== "") { 
			$hppId = $this->options['production_hpp_id'];
		}

		if ( !isset( $hppId ) ) {
			$this->_logger->info( 'The ID of the Hosted Payment Page is not set in WP Settings' );
			return false;
		}

		if ( !in_array( strlen( trim($hppId) ), [ 24, 36 ], true ) ) { 
			$this->_logger->info( 'The ID of the Hosted Payment Page is set, but of an invalid length. Must be 24 or 36 characters in length.' );
			return false;
		}

		return $hppId;
	}

	/** 
	 * Get User ID based on the currently configured environmet
	 * 
	 * @return String|Bool $userId|false
	 */
	private function getUserId() {
		if (trim($this->options['environment']) === "sandbox" && trim($this->options['sandbox_user_id']) !== "") { 
			$userId = $this->options['sandbox_user_id'];
		}

		if (trim($this->options['environment']) === "production" && trim($this->options['production_user_id']) !== "") { 
			$userId = $this->options['production_user_id'];
		}

		if ( !isset( $userId ) ) {
			$this->_logger->info( 'User ID is not set in WC Settings' );
			return false;
		}

		if ( !in_array( strlen( trim($userId) ), [ 24, 36 ], true ) ) { 
			$this->_logger->info( 'User ID is of an invalid length. Must be 24 or 36 charcters in length' );
			return false;
		}

		return $userId;
	}
	
	/** 
	 * Get User API Key based on the currently configured environmet
	 * 
	 * @return String|Bool $userApiKey|false
	 */
	private function getUserApiKey() {
		if (trim($this->options['environment']) === "sandbox" && trim($this->options['sandbox_user_api_key']) !== "") { 
			$userApiKey = $this->options['sandbox_user_api_key'];
		}

		if (trim($this->options['environment']) === "production" && trim($this->options['production_user_api_key']) !== "") { 
			$userApiKey = $this->options['production_user_api_key'];
		}

		if ( !isset( $userApiKey ) ) {
			$this->_logger->info( 'API Key is not set in WP Settings' );
			return false;
		}

		if ( strlen(trim($userApiKey)) < 1 && strlen(trim($userApiKey)) > 64 ) {
			$this->_logger->info( 'API Key is input but set at an invalid length. API Key myst be > 1 character and <= 64 characters' );
			return false;
		}

		return $userApiKey;
	}

	private function getItemTaxRate($item) { 
		$_tax = new WC_Tax();
		$product_tax_class = $item->get_tax_class();
		$tax = $_tax->get_rates($product_tax_class); // get tax rate array for item

		$taxesArray = array_keys($tax); // get array keys as $key is the id of the tax rate
		$taxRateId = $taxesArray[0]; // get the first item of the array above, to extract the tax_rate_id
		$itemTaxRate = $tax[$taxRateId]["rate"]; // get the rate of the above it.
		return $itemTaxRate;
	}

	/** 
	 * Make API Call to Host API
	 * 
	 * @return Array $result | Result from cURL Request
	 */
	private function makeApiCall( $url, $method, $payload = "" ) {
		$this->_logger->info( 'Making '. $method .' request to the endpoint: '. $url );
		$this->_logger->info( 'Payload: '. json_encode($payload) );

		$ch = curl_init();
		$header = array('Content-Type: application/json');

		if($this->getEnvironment() === "sandbox") { 
			array_push($header, 'developer-id: '. 'ei5moHkT');
		} else {
			array_push($header, 'developer-id: '. 'Woocom21');
		}

		array_push($header, 'user-id: '. $this->getUserId()); 

		$this->_logger->info("Line: " . __LINE__ . " User ID being used: " . $this->getUserId());

        array_push($header, 'user-api-key: '. $this->getUserApiKey()); 

        $curlConfig = array(
            CURLOPT_HTTPHEADER     => $header,
            CURLOPT_URL            => $url,
            CURLOPT_RETURNTRANSFER => true
		);

        if ( isset( $method ) ) {
            if ( strtolower( $method ) == 'post' ) {
                $curlConfig[ CURLOPT_POST ] = true;
            }
            if ( strtolower( $method ) == 'get' ) {
                $curlConfig[ CURLOPT_HTTPGET ] = true;
			}
			if ( strtolower($method ) == 'put') {
                $curlConfig[ CURLOPT_CUSTOMREQUEST ] = "PUT";
			}	
		}

        if ( $payload !== "" ) {
            $curlConfig[CURLOPT_POSTFIELDS] = json_encode( $payload );
		}
		
        curl_setopt_array( $ch, $curlConfig );
		$result = curl_exec( $ch );
		curl_close( $ch );
		
		$this->_logger->info("Line: " . __LINE__ . " API Response: " . json_encode($result));

        if ( !$result ) {
			$this->_logger->info( 'Bad API Call: Result ' . json_encode( $result ) );
            return $result;
        }

        return json_decode( $result, true );
	}
}